/* pronto - regexp search for PROSITE patterns
 *          uses the GNU regexp library, regex 0.12
 *
 * SRE, Wed Oct 26 16:10:03 1994
 */
#include <stdio.h>
#include <string.h>
#include "squid.h"		/* my biosequence library */
#include "regex.h"		/* GNU regex library      */

#define MAX_REGEX_LEN 512
#define OPTIONS "i"

static float 
rfsfq[20] = { .081, .015, .054, .061, .040, .068, .022, .057, .056, .093,
              .025, .045, .049, .039, .057, .068, .058, .067, .013, .032};

char usage[]  = "Usage: pronto [-options] <PROSITE pattern> <seqfile>\n\
or: pronto -i <PROSITE pattern>\n\
  where available options are:\n\
  -i   : print out negative log probability of match only\n";

char banner[] = "pronto -- search for PROSITE matches in sequence(s)\n\
     version 1.0, SRE, August 1993";

static int prosite2regex(char *prosite, char *regex);
static float prosite_info(char *prosite);

int
main(int argc, char **argv)
{
  char  *seqfile;               /* name of sequence file */
  SQFILE *dbfp;			/* open sequence file */
  int    fmt;			/* format of seqfile  */
  char  *seq;			/* sequence */
  SQINFO sqinfo;                /* additional sequence info */
  char  *prosite;               /* prosite pattern */
  char   regex[512];    	/* regular expression */
  int    regex_len;		/* length of regular expression  */
  struct re_pattern_buffer regex_comp; /* compiled regexp */
  char  *re_err;		/* error msg returned by regex */
  int    start;			/* start for regexp search */
  int    range;			/* range of regexp search */
  struct re_registers regs;	/* start, end of match  */
  int    info_only;

  int    optchar;		/* option character, command line */
  extern int   optind;
  extern char *optarg;

  /***********************************************
   * Parse command line
   ***********************************************/

  re_syntax_options = RE_SYNTAX_POSIX_EGREP;
  info_only         = FALSE;

  while ((optchar = getopt(argc, argv, OPTIONS)) != -1)
    switch (optchar) {

    case 'i': info_only = TRUE; break;
    default:
      Die("Unrecognized option: -%c\n%s\n", optchar, usage);
    }

  if (info_only == FALSE && argc - optind != 2)
    Die("%s\n", usage); 
  if (info_only == TRUE && argc - optind != 1)
    Die("%s\n", usage); 

  prosite = argv[optind]; optind++;
  if (info_only == FALSE) seqfile = argv[optind];
		 
  if (info_only) 
    { 
      printf("%.2f\n", prosite_info(prosite));
      exit(0);
    }

  prosite2regex(prosite, regex);

  if (! SeqfileFormat(seqfile, &fmt, NULL))
    Die("Failed to determine format of file %s", seqfile);

  if ((dbfp = SeqfileOpen(seqfile, fmt, NULL)) == NULL)
    fprintf(stderr, "Failed to open sequence file %s for reading", seqfile);
  
  

  regex_comp.translate = 0;
  regex_comp.buffer    = 0;
  regex_comp.allocated = 0;
  regex_comp.fastmap   = (char *) MallocOrDie(256);
  regex_len            = strlen(regex);
  if ((re_err = (char *) re_compile_pattern(regex, regex_len, &regex_comp)) != NULL)
    Die(re_err);

  while (ReadSeq(dbfp, fmt, &seq, &sqinfo))
    {
      start = 0;
      range = sqinfo.len;

      while (re_search(&regex_comp, seq, sqinfo.len, start, range, &regs) >= 0)
	{
	  printf("%6d  %6d   %s %s\n",
		 regs.start[0]+1, regs.end[0],
		 sqinfo.name,
		 (sqinfo.flags & SQINFO_DESC) ? sqinfo.desc : "");
	  start = regs.end[0];
	}
      FreeSequence(seq, &sqinfo);
    }

  regfree(&regex_comp);
  SeqfileClose(dbfp);
  return 0;
}

static int
prosite2regex(char *prosite, char *regex)
{
  char *regptr;

  regptr = regex;
  for (; *prosite != '.' && *prosite != '\0'; prosite++)
    {
      if (strchr(AMINO_ALPHABET, *prosite) != NULL) 
	*regptr++ = *prosite;

      else if (*prosite == 'x' || *prosite == 'X') 
	*regptr++ = '.'; 

      else if (*prosite == '[' || *prosite == ']')
	*regptr++ = *prosite;

      else if (*prosite == '-')
	continue;

      /* Rules for negation
       */
      else if (*prosite == '{')
	{ 
	  *regptr++ = '[';
	  *regptr++ = '^';
	}
      else if (*prosite == '}')
	*regptr++ = ']';

      
      /* Rules for ranges
       */
      else if (*prosite == '(')
	*regptr++ = '{';
      else if (isdigit(*prosite) || *prosite == ',')
	*regptr++ = *prosite;
      else if (*prosite == ')')
	*regptr++ = '}';

      /* Begin-sequence and end-sequence anchor rules
       */
      else if (*prosite == '<')
	*regptr++ = '^';
      else if (*prosite == '>')
	*regptr++ = '$';

      else
	Die("Can't deal with %c in PROSITE pattern %s", *prosite, prosite);
    }
  *regptr = '\0';
  return 1;
}

/* Calculate the negative log probability of a PROSITE match,
 * in base two (bits). Uses the Staden approximation. See Roger
 * Sewell's paper on calculating Prosite match probabilities.
 */
static float
prosite_info(char *prosite)
{
  float  info;
  char  *sptr;
  float  lastatom;		/* probability of last atom */
  float  gapsize;
  float  lowrange;
  int    negatom;		/* true if a negation atom  */
  char   alphabet[21] = "ACDEFGHIKLMNPQRSTVWY";

  info = lastatom = 0.0;
  gapsize  = 1.0;
  lowrange = 1.0;
  negatom  = FALSE;
  for (; *prosite != '.' && *prosite != '\0'; prosite++)
    {
      if ((sptr = strchr(alphabet, *prosite)) != NULL) 
	lastatom += rfsfq[sptr-alphabet];

      else if (*prosite == 'x' || *prosite == 'X') 
	lastatom = 1.0;

      else if (*prosite == '[' || *prosite == ']')
	negatom = FALSE;

      else if (*prosite == '-')
	{			/* collect previous atom probability */
	  if (negatom) lastatom = 1.0 - lastatom;
	  info += log(lastatom) * lowrange;
	  info += log(gapsize);
	  lastatom = 0.0;
	  lowrange = 1.0;
	  gapsize = 1.0;
	  negatom = FALSE;
	}

      /* Rules for negation
       */
      else if (*prosite == '{' || *prosite == '}')
	negatom = TRUE;

      /* Rules for ranges
       */
      else if (*prosite == '(')
	{
	  sptr = prosite+1;
	  prosite = strchr(prosite, ')');
	  *prosite = '\0';

	  lowrange = (float) atoi(sptr);
	  if ((sptr = (strchr(sptr, ','))) != NULL)
	    gapsize = (float) atoi(sptr+1) - lowrange + 1.0;
	}

      /* Begin-sequence and end-sequence anchor rules;
       * ignored for info content effect
       */
      else if (*prosite == '<')
	continue;
      else if (*prosite == '>')
	continue;

      else
	Die("Can't deal with %c in PROSITE pattern %s", *prosite, prosite);
    }

  if (negatom) lastatom = 1.0 - lastatom;
  info += log(lastatom) * lowrange;
  info += log(gapsize);
  return (info * -1.442695);	/* returns negative entropy in bits */
  return ;
}

