#! /usr/local/bin/perl

# Usage: prosite_search.pl <cosmid name>
#
# Searches a cosmid for matches to the PROSITE motif database
# and creates .ace format data file for ACEDB.
# Uses only patterns; rules and profiles are excluded from the search.
#
# Input:    cosmid DNA sequence, in FASTA format, in $cosmid.seq 
# Output:   ace format
#
# Environment variables:
#   PROSITEDIR must be set. Expects to find $PROSITEDIR/prosite.dat
#
# Temporary files:
#   Temporarily creates "sre_orfs.tmp".
#
# Dependencies:
#   "translate", "pronto", and "seqstat" must be in the $path.
# 
# Options:
#    -i <info_cutoff>  : only search with patterns with more
#                        than this negative log P, in bits 

require "importenv.pl";
require "getopts.pl";

&Getopts('i:');
if ($opt_i) { $infocutoff = $opt_i; }
$cosmid  = shift;
$seqfile = "$cosmid.seq";
$orffile = "sre_orfs.tmp";

$usage = 
"prosite_search.pl [-options] <cosmid name>\n\
   where available options are:\n\
   -i <info_cutoff>  : only search with patterns with more\n\
                       than this negative log P, in bits\n";

die ("No cosmid sequence file $cosmid.seq; died") unless -e "$cosmid.seq";


# Translate the cosmid in all six frames; get its length.
#
system ("translate -a $seqfile > $orffile");
$seqlen = &single_seqlen($seqfile);

$/ = "//";			# Paragraph mode on // separators.
open(PROSITE,"$PROSITEDIR/prosite.dat") || die("Failed to open PROSITEDIR/prosite.dat;\n perhaps you haven't set PROSITEDIR? Died");

while (<PROSITE>) {
    @lines = split(/^/);
    while ($line = shift(@lines)) {
	if ($line =~ /^ID\s+(\S+); (\S+)/) { $name = $1; $patrule = $2; }
	if ($line =~ /^AC\s+(\S+);/) { $acc = $1; }
	if ($line =~ /^DE\s+(.+)$/)  { $desc = $desc.$1; }
	if ($line =~ /^PA\s+(\S+)\s*$/) { $pattern = $pattern.$1; }
    }
    if ($patrule eq "PATTERN.") {
	$info = `pronto -i \"$pattern\"`;
	chop ($info);
	if ($info > $infocutoff) {
	    $output = `pronto \"$pattern\" $orffile`;


	    # Parse the output.
	    @outlines = split(/^/, $output);
	    while ($outline = shift(@outlines)) {
		($aafrom, $aato, $frame) = ($outline =~ /^\s*(\d+)\s+(\d+)\s+\S+:(\d)\s*/);

		$hitlen = $aato - $aafrom + 1;

		if ($frame < 3) { 
		    $ntfrom = ($aafrom - 1) * 3 + $frame + 1;
		    $ntto   = ($aato - 1) * 3   + $frame + 3;
		}
		else {
		    $ntfrom = $seqlen - (($aafrom - 1) * 3 + $frame - 3);
		    $ntto   = $seqlen - (($aato - 1) * 3 + $frame - 1);
		}

		# Print the ACE entry for the cosmid
		print "Sequence $cosmid\n";
		print "Homol $acc Prosite $info $ntfrom $ntto 1 $hitlen\n\n";

		# Print the ACE entry for the prosite pattern
		print "Sequence : $acc\n";
		print "Title \"$desc\"\n";
		print "Library PROSITE $acc $acc\n";
		print "Peptide\n";
		print "Homol $cosmid Prosite $info 1 $hitlen $ntfrom $ntto\n\n";
	    }
	}
	$patrule = 0;
	$pattern = "";
	$desc = "";
    }
}

# Clean up.
#
unlink ($orffile);


sub
single_seqlen {
    local($seqfile) = @_;
    local($output);

    $output = `seqstat $seqfile`;
    if ($output =~ /(\d+) sequences;\s*(\d+)\s*residues/) {
	$nseq   = $1;
	$seqlen = $2;
    }
    else { die "Failed to parse seqstat output; died"; }
    die "$cosmid file has > 1 seq in it" unless $nseq < 2;

    $seqlen;
}
